#!/usr/bin/env python
from __future__ import absolute_import

import sys


def main():
    from pganonymize.cli import get_arg_parser, main

    try:
        args = get_arg_parser().parse_args()
        main(args)
        exit_status = 0
    except KeyboardInterrupt:
        exit_status = 1
    sys.exit(exit_status)


if __name__ == '__main__':
    main()



"""Helper methods"""

from __future__ import absolute_import

import json
import logging
import math
import re
import subprocess
import time

import parmap
import psycopg2
import psycopg2.extras
from pgcopy import CopyManager
from psycopg2.sql import SQL, Composed, Identifier
from tqdm import trange

from pganonymize.config import config
from pganonymize.constants import DEFAULT_CHUNK_SIZE, DEFAULT_PRIMARY_KEY
from pganonymize.providers import provider_registry

# Needed to work with UUID objects
psycopg2.extras.register_uuid()


def anonymize_tables(connection, verbose=False, dry_run=False):
    """
    Anonymize a list of tables according to the schema definition.

    :param connection: A database connection instance.
    :param bool verbose: Display logging information and a progress bar.
    :param bool dry_run: Script is running in dry-run mode, no commit expected.
    """
    definitions = config.schema.get('tables', [])
    for definition in definitions:
        start_time = time.time()
        table_name = list(definition.keys())[0]
        logging.info('Found table definition "%s"', table_name)
        table_definition = definition[table_name]
        columns = table_definition.get('fields', [])
        excludes = table_definition.get('excludes', [])
        search = table_definition.get('search')
        primary_key = table_definition.get('primary_key', DEFAULT_PRIMARY_KEY)
        total_count = get_table_count(connection, table_name, dry_run)
        chunk_size = table_definition.get('chunk_size', DEFAULT_CHUNK_SIZE)
        build_and_then_import_data(connection, table_name, primary_key, columns, excludes,
                                   search, total_count, chunk_size, verbose=verbose, dry_run=dry_run)
        end_time = time.time()
        logging.info('{} anonymization took {:.2f}s'.format(table_name, end_time - start_time))


def process_row(row, columns, excludes):
    if row_matches_excludes(row, excludes):
        return None
    else:
        row_column_dict = get_column_values(row, columns)
        if row_column_dict:
            for key, value in row_column_dict.items():
                row[key] = value
        else:
            return None
        return row


def build_and_then_import_data(connection, table, primary_key, columns,
                               excludes, search, total_count, chunk_size, verbose=False, dry_run=False):
    """
    Select all data from a table and return it together with a list of table columns.

    :param connection: A database connection instance.
    :param str table: Name of the table to retrieve the data.
    :param str primary_key: Table primary key
    :param list columns: A list of table fields
    :param list[dict] excludes: A list of exclude definitions.
    :param str search: A SQL WHERE (search_condition) to filter and keep only the searched rows.
    :param int total_count: The amount of rows for the current table
    :param int chunk_size: Number of data rows to fetch with the cursor
    :param bool verbose: Display logging information and a progress bar.
    :param bool dry_run: Script is running in dry-run mode, no commit expected.
    """
    column_names = get_column_names(columns)
    sql_columns = SQL(', ').join([Identifier(column_name) for column_name in [primary_key] + column_names])
    sql_select = SQL('SELECT {columns} FROM {table}').format(table=Identifier(table), columns=sql_columns)
    if search:
        sql_select = Composed([sql_select, SQL(" WHERE {search_condition}".format(search_condition=search))])
    if dry_run:
        sql_select = Composed([sql_select, SQL(" LIMIT 100")])
        logging.info(sql_select.as_string(connection))
    cursor = connection.cursor(cursor_factory=psycopg2.extras.DictCursor, name='fetch_large_result')
    cursor.execute(sql_select.as_string(connection))
    temp_table = 'tmp_{table}'.format(table=table)
    create_temporary_table(connection, columns, table, temp_table, primary_key)
    batches = int(math.ceil((1.0 * total_count) / (1.0 * chunk_size)))
    for i in trange(batches, desc="Processing {} batches for {}".format(batches, table), disable=not verbose):
        records = cursor.fetchmany(size=chunk_size)
        if records:
            data = parmap.map(process_row, records, columns, excludes, pm_pbar=verbose)
            import_data(connection, temp_table, [primary_key] + column_names, filter(None, data))
    apply_anonymized_data(connection, temp_table, table, primary_key, columns)

    cursor.close()


def apply_anonymized_data(connection, temp_table, source_table, primary_key, definitions):
    logging.info('Applying changes on table {}'.format(source_table))
    cursor = connection.cursor()
    create_index_sql = SQL('CREATE INDEX ON {temp_table} ({primary_key})')
    sql = create_index_sql.format(temp_table=Identifier(temp_table), primary_key=Identifier(primary_key))
    cursor.execute(sql.as_string(connection))

    column_names = get_column_names(definitions)
    columns_identifiers = [SQL('{column} = s.{column}').format(column=Identifier(column)) for column in column_names]
    set_columns = SQL(', ').join(columns_identifiers)
    sql_args = {
        'table': Identifier(source_table),
        'columns': set_columns,
        'source': Identifier(temp_table),
        'primary_key': Identifier(primary_key)
    }
    sql = SQL(
        'UPDATE {table} t '
        'SET {columns} '
        'FROM {source} s '
        'WHERE t.{primary_key} = s.{primary_key}'
    ).format(**sql_args)
    cursor.execute(sql.as_string(connection))
    cursor.close()


def row_matches_excludes(row, excludes=None):
    """
    Check whether a row matches a list of field exclusion patterns.

    :param list row: The data row
    :param list excludes: A list of field exclusion roles, e.g.:

    >>> [
    >>>     {'email': ['\\S.*@example.com', '\\S.*@foobar.com', ]}
    >>> ]

    :return: True or False
    :rtype: bool
    """
    excludes = excludes if excludes else []
    for definition in excludes:
        column = list(definition.keys())[0]
        for exclude in definition.get(column, []):
            pattern = re.compile(exclude, re.IGNORECASE)
            if row[column] is not None and pattern.match(row[column]):
                return True
    return False


def create_temporary_table(connection, definitions, source_table, temp_table, primary_key):
    primary_key = primary_key if primary_key else DEFAULT_PRIMARY_KEY
    column_names = get_column_names(definitions)
    sql_columns = SQL(', ').join([Identifier(column_name) for column_name in [primary_key] + column_names])
    ctas_query = SQL("""CREATE TEMP TABLE {temp_table} AS SELECT {columns}
                    FROM {source_table} WITH NO DATA""")
    cursor = connection.cursor()
    cursor.execute(ctas_query.format(temp_table=Identifier(temp_table),
                                     source_table=Identifier(source_table), columns=sql_columns)
                   .as_string(connection)
                   )
    cursor.close()


def import_data(connection, table_name, column_names, data):
    """
    Import the temporary and anonymized data to a temporary table and write the changes back.

    :param connection: A database connection instance.
    :param str table_name: Name of the table to be populated with data.
    :param list column_names: A list of table fields
    :param list data: The table data.
    """
    mgr = CopyManager(connection, table_name, column_names)
    mgr.copy([[escape_str_replace(val) for col, val in row.items()] for row in data])


def get_connection(pg_args):
    """
    Return a connection to the database.

    :param pg_args:
    :return: A psycopg connection instance
    :rtype: psycopg2.connection
    """
    return psycopg2.connect(**pg_args)


def get_table_count(connection, table, dry_run):
    """
    Return the number of table entries.

    :param connection: A database connection instance
    :param str table: Name of the database table
    :param bool dry_run: Script is running in dry-run mode, no commit expected.
    :return: The number of table entries
    :rtype: int
    """
    if dry_run:
        return 100
    else:
        sql = SQL('SELECT COUNT(*) FROM {table}').format(table=Identifier(table))
        cursor = connection.cursor()
        cursor.execute(sql.as_string(connection))
        total_count = cursor.fetchone()[0]
        cursor.close()
        return total_count


def get_column_values(row, columns):
    """
    Return a dictionary for a single data row, with altered data.

    :param psycopg2.extras.DictRow row: A data row from the current table to be altered
    :param list columns: A list of table columns with their provider rules, e.g.:

    >>> [
    >>>     {'guest_email': {'append': '@localhost', 'provider': 'md5'}}
    >>> ]

    :return: A dictionary with all fields that have to be altered and their value for a single data row, e.g.:
        {'guest_email': '12faf5a9bb6f6f067608dca3027c8fcb@localhost'}
    :rtype: dict
    """
    column_dict = {}
    for definition in columns:
        full_column_name = get_column_name(definition, True)
        column_name = get_column_name(definition, False)
        column_definition = definition[full_column_name]
        provider_config = column_definition.get('provider')
        orig_value = nested_get(row, full_column_name)
        # Skip the current column if there is no value to be altered
        if orig_value is not None:
            provider_class = provider_registry.get_provider(provider_config['name'])
            value = provider_class.alter_value(orig_value, **provider_config)
            append = column_definition.get('append')
            if append:
                value = value + append
            _format = column_definition.get('format')
            if _format:
                value = _format.format(pga_value=value, **row)
            nested_set(row, full_column_name, value)
            column_dict[column_name] = nested_get(row, column_name)
    return column_dict


def truncate_tables(connection):
    """
    Truncate a list of tables.

    :param connection: A database connection instance
    """
    tables = config.schema.get('truncate', [])
    if not tables:
        return
    cursor = connection.cursor()
    table_names = SQL(', ').join([Identifier(table_name) for table_name in tables])
    logging.info('Truncating tables "%s"', table_names)
    cursor.execute(SQL('TRUNCATE TABLE {tables}').format(tables=table_names).as_string(connection))
    cursor.close()


def create_database_dump(filename, db_args):
    """
    Create a dump file from the current database.

    :param str filename: Path to the dumpfile that should be created
    :param dict db_args: A dictionary with database related information
    """
    arguments = '-d {dbname} -U {user} -h {host} -p {port}'.format(**db_args)
    cmd = 'pg_dump -Fc -Z 9 {args} -f {filename}'.format(
        args=arguments,
        filename=filename
    )
    logging.info('Creating database dump file "%s"', filename)
    subprocess.call(cmd, shell=True)


def get_column_name(definition, fully_qualified=False):
    """
    Get column name by definition.

    :param dict definition: Column definition
    :param bool fully_qualified: Get complete column name with path (json objects)
    :return: A string, containing column name. ex:
        id
        name
        metadata.col1
    :rtype: string
    """
    col_name = list(definition.keys())[0]
    if fully_qualified:
        return col_name
    else:
        return col_name.split('.', 2)[0]


def get_column_names(definitions):
    """
    Get distinct column names from definitions

    :param list definitions: A list of table definitions from the YAML schema.
    :return: A list of column names
    :rtype: list
    """
    names = []
    for definition in definitions:
        name = get_column_name(definition)
        if name not in names:
            names.append(name)
    return names


def escape_str_replace(value):
    """
    Get escaped value.

    :param value: The value to be encoded.
    :return: Escaped value
    """
    if isinstance(value, dict):
        return json.dumps(value, default=str).encode()
    return value


def nested_get(dic, path, delimiter='.'):
    """
    Get from dictionary by path.

    :param dict dic: The source dictionary.
    :param str path: The path within the dictionary.
    :param str delimiter: The path delimiter
    :return: Value at path
    """
    try:
        keys = path.split(delimiter)
        for key in keys[:-1]:
            dic = dic.get(key, {})
        return dic[keys[-1]]
    except (AttributeError, KeyError, TypeError):
        return None


def nested_set(dic, path, value, delimiter='.'):
    """
    Set dictionary value by path.

    :param dict dic: The source dictionary
    :param str path: The path withing dictionary
    :param value: The value to be set
    :param str delimiter: The path delimiter
    """
    keys = path.split(delimiter)
    for key in keys[:-1]:
        dic = dic.get(key, {})
    dic[keys[-1]] = value



# Database arguments used for the CLI
DATABASE_ARGS = ('dbname', 'user', 'password', 'host', 'port')

# Default name for the primary key column
DEFAULT_PRIMARY_KEY = 'id'

# Filename of the default schema
DEFAULT_SCHEMA_FILE = 'schema.yml'

# Default chunk size for data fetch
DEFAULT_CHUNK_SIZE = 100000



import operator
import random
import re
from collections import OrderedDict
from hashlib import md5
from uuid import uuid4

from faker import Faker

from pganonymize.config import config
from pganonymize.exceptions import InvalidProvider, InvalidProviderArgument, ProviderAlreadyRegistered


class FakerInitializer(object):
    """A wrapper that allows to instantiate a faker instance with specific locales."""

    def __init__(self):
        self._faker = None
        self._options = None

    @property
    def options(self):
        if self._options is None:
            self._options = config.schema.get('options', {}).get('faker', {})
        return self._options

    @property
    def default_locale(self):
        return self.options.get('default_locale')

    @property
    def faker(self):
        """
        Return the actual :class:`faker.Faker` instance, with optional locales taken from the YAML schema.

        :return: A faker instance
        :rtype: faker.Faker
        """
        if self._faker is None:
            locales = self.options.get('locales')
            self._faker = Faker(locales)
        return self._faker

    def get_locale_generator(self, locale):
        """
        Get the internal generator for the given locale.

        :param str locale: A locale string
        :raises InvalidProviderArgument: If locale is unknown (not configured within the global locales option).
        :return: A Generator instance for the given locale
        :rtype: faker.Generator
        """
        try:
            generator = self.faker[locale]
        except KeyError:
            raise InvalidProviderArgument('Locale \'{}\' is unknown. Have you added it to the global option '
                                          '(options.faker.locales)?'.format(locale))
        return generator


faker_initializer = FakerInitializer()


class ProviderRegistry(object):
    """A registry for provider classes."""

    def __init__(self):
        self._registry = OrderedDict()

    def register(self, provider_class, provider_id):
        """
        Register a provider class.

        :param pganonymize.providers.Provider provider_class: Provider class that should be registered
        :param str provider_id: A string id to register the provider for
        :raises ProviderAlreadyRegistered: If another provider with the given id has been registered
        """
        if provider_id in self._registry:
            raise ProviderAlreadyRegistered('A provider with the id "{}" has already been registered'
                                            .format(provider_id))
        self._registry[provider_id] = provider_class

    def get_provider(self, provider_id):
        """
        Return a provider by its provider id.

        :param str provider_id: The string id of the desired provider.
        :raises InvalidProvider: If no provider can be found with the given id.
        :return: The provider class that matches the id.
        :rtype: type
        """
        for key, cls in self._registry.items():
            if (cls.regex_match is True and re.match(re.compile(key), provider_id) is not None) or key == provider_id:
                return cls
        raise InvalidProvider('Could not find provider with id "{}"'.format(provider_id))

    @property
    def providers(self):
        """
        Return the registered providers.

        :rtype: OrderedDict
        """
        return self._registry


provider_registry = ProviderRegistry()


def register(provider_id, **kwargs):
    """
    A wrapper that registers a provider class to the provider registry.

    :param str provider_id: The string id to register the provider for.
    :keyword registry: The registry the provider class is registered at (default is the `provider_registry` instance).
    :return: The decorator function
    :rtype: function
    """
    def wrapper(provider_class):
        registry = kwargs.get('registry', provider_registry)
        registry.register(provider_class, provider_id)
        return provider_class
    return wrapper


class Provider(object):
    """Base class for all providers."""

    regex_match = False
    """Defines whether a provider matches it's id using regular expressions."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        """
        Alter or replace the original value of the database column.

        :param original_value: The original value of the database column.
        """
        raise NotImplementedError()


@register('choice')
class ChoiceProvider(Provider):
    """Provider that returns a random value from a list of choices."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        return random.choice(kwargs.get('values'))


@register('clear')
class ClearProvider(Provider):
    """Provider to set a field value to None."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        return None


@register('fake.+')
class FakeProvider(Provider):
    """Provider to generate fake data."""

    regex_match = True

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        func_name = kwargs['name'].split('.', 1)[1]
        func_kwargs = kwargs.get('kwargs', {})
        locale = kwargs.get('locale', faker_initializer.default_locale)
        # Use the generator for the locale if a locale is configured (per field definition or as global default locale)
        faker_generator = faker_initializer.get_locale_generator(locale) if locale else faker_initializer.faker
        try:
            func = operator.attrgetter(func_name)(faker_generator)
        except AttributeError as exc:
            raise InvalidProviderArgument(exc)
        return func(**func_kwargs)


@register('mask')
class MaskProvider(Provider):
    """Provider that masks the original value."""

    default_sign = 'X'
    """The default string used to replace each character."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        sign = kwargs.get('sign', cls.default_sign) or cls.default_sign
        return sign * len(original_value)


@register('partial_mask')
class PartialMaskProvider(Provider):
    """Provider that masks some of the original value."""

    default_sign = 'X'
    default_unmasked_left = 1
    default_unmasked_right = 1
    """The default string used to replace each character."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        sign = kwargs.get('sign', cls.default_sign) or cls.default_sign
        unmasked_left = kwargs.get('unmasked_left', cls.default_unmasked_left) or cls.default_unmasked_left
        unmasked_right = kwargs.get('unmasked_right', cls.default_unmasked_right) or cls.default_unmasked_right

        return (
            original_value[:unmasked_left] +
            (len(original_value) - (unmasked_left + unmasked_right)) * sign +
            original_value[-unmasked_right:]
        )


@register('md5')
class MD5Provider(Provider):
    """Provider to hash a value with the md5 algorithm."""

    default_max_length = 8
    """The default length used for the number representation."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        as_number = kwargs.get('as_number', False)
        as_number_length = kwargs.get('as_number_length', cls.default_max_length)
        hashed = md5(original_value.encode('utf-8')).hexdigest()
        if as_number:
            return int(hashed, 16) % (10 ** as_number_length)
        else:
            return hashed


@register('set')
class SetProvider(Provider):
    """Provider to set a static value."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        return kwargs.get('value')


@register('uuid4')
class UUID4Provider(Provider):
    """Provider to set a random uuid value."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        return uuid4()


@register('update_json')
class UpdateJSONProvider(Provider):
    """Provider to update JSON data (currently values) by providers."""

    @classmethod
    def alter_value(cls, original_value, **kwargs):
        def update_dict(input_dict, update_values_type={}):
            """Update dictionary with recursion (nested dictionaries)."""
            if not update_values_type:
                return
            for key, val in input_dict.items():
                if isinstance(val, dict):
                    update_dict(val, update_values_type=update_values_type)
                else:
                    val_type = type(val).__name__
                    val_update = update_values_type.get(val_type)
                    if val_update:
                        if val_update.get('provider'):
                            provider_config = val_update.get('provider')
                            provider_class = provider_registry.get_provider(provider_config['name'])
                            provider_value = provider_class.alter_value(val, **provider_config)
                            input_dict[key] = provider_value
        update_dict(original_value, update_values_type=kwargs.get('update_values_type', {}))
        return original_value



import os
import re

import yaml


class Config(object):
    """A class that wraps access to the given YAML schema file."""

    def __init__(self):
        self._schema = None
        self.schema_file = None

    @property
    def schema(self):
        """
        Return the schema loaded from the given YAML schema file.

        :return: The parsed YAML schema.
        :rtype: dict
        """
        if self._schema is None and self.schema_file is not None:
            self._schema = load_schema(self.schema_file)
        return self._schema


def load_schema(schema_file):
    # Original code from here https://gist.github.com/mkaranasou/ba83e25c835a8f7629e34dd7ede01931
    tag = '!ENV'
    pattern = re.compile(r'.*?\${(\w+)}.*?')
    custom_loader = yaml.FullLoader
    custom_loader.add_implicit_resolver(tag, pattern, None)

    def constructor_env_variables(loader, node):
        """
        Extract the environment variable from the node's value.

        :param yaml.Loader loader: The yaml loader
        :param node: The current node in the yaml
        :return: The parsed string that contains the value of the environment variable
        """
        value = loader.construct_scalar(node)
        match = pattern.findall(value)  # to find all env variables in line
        if match:
            full_value = value
            for g in match:
                full_value = full_value.replace(
                    '${{{g}}}'.format(g=g), os.environ.get(g, g)
                )
            return full_value
        return value

    custom_loader.add_constructor(tag, constructor_env_variables)
    return yaml.load(open(schema_file), Loader=custom_loader)


config = Config()
